#include "main.h"


void moves_to_pos (char *fen, board * B)
{
    int mv;
    int prom = 0; //  - ?

    while (*fen)
    {
        while (*fen == ' ') fen ++;

        if (!(*fen)) return;

        mv = str_to_move (fen);
        if (!mv) return;

        if (mv & f_prom)
        {
            prom = 1;
        };

        if (!do_move (mv, prom, B))
        {
            uci_send ("info string Can't understand move line\n");

            print_board (B);

            uci_send ("%s\n", fen);

            return;
        };

        while (*fen != ' ' && *fen) fen ++;
        //fen += 4; //     4 !
    };
};

char* read_command (char *str, const char *wrd)
{
	for (;; str++)
	{
		if (!*str) return NULL;
		if (*str == *wrd) break;
	};

	for (;; str++, wrd++)
	{
	    if (!*wrd) return str; //  ,    - 
	    if (!*str || !*wrd) return NULL;
		if (*str != *wrd) return NULL;
	};

	return str;
};

int read_number (char *str)
{
	char ch;
	int num = 0;

	for (;; str ++)
	{
		ch = *str;
		if (ch < '0' || ch > '9') break;

		num *= 10;
		num += ch - '0';
	};

	return num;
};

char* move_to_str (int mv)
{
    static char buff [6];
    int _flags, _to, _from;

    _flags = mv & 0xFF;
    mv >>= 8;

    _to    = mv & 0xFF;
    mv >>= 8;

    _from  = mv & 0xFF;
    mv >>= 8;

    sprintf (buff    , "%c%c", 'a' + (vert (_from)), '1' + (hori (_from)));
    sprintf (buff + 2, "%c%c", 'a' + (vert (_to))  , '1' + (hori (_to)));

    if (is_prom (_flags)) sprintf (buff + 4, "%c\0", "nbrq"[_flags & ~f_mask]);
    else                  sprintf (buff + 4, "\0");

    return buff;
}

int str_to_move (char * s)
{
    char a, b, c, d, e;
    int _from, _to, _flags;

    a = s[0] - 'a';
    b = s[1] - '1';
    c = s[2] - 'a';
    d = s[3] - '1';
    e = s[4];

    //uci_send ("prepared to promote: %c\n", e);

    _flags = 0;

    switch (e)
    {
        case 'n': _flags = f_prom + 0; break;
        case 'b': _flags = f_prom + 1; break;
        case 'r': _flags = f_prom + 2; break;
        case 'q': _flags = f_prom + 3; break;
        case ' ': _flags = 0; break;
        default : _flags = 0; break; //   - 
    };

    if (a < 0 || a > 7) return 0;
    if (b < 0 || b > 7) return 0;
    if (c < 0 || c > 7) return 0;
    if (d < 0 || d > 7) return 0;

    _from = compose (b, a);
    _to   = compose (d, c);

    //if ((b == d) && (absu (a, c) == 2)) _flags |= f_castling;

    return (_from << 16) | (_to << 8) | _flags;
};

void init_pst ()
{
    int a, b, i;
    int sq, t;
    U64 att;
    U64 x64; //  popcount'
    const int pawn_pst_x [] = {-3, -1, 0, 1, 1, 0, -1, -3};
    const int pawn_pst_y [] = {0, -10, -5, 0, 2, 4, 10, 0};
    const int knight_pst [] = {0, 0, -20, -10, -5, 0, 5, 0, 8};
    const int bishop_pst [] = {0, 0, 0, 0, 0, 0, 0, -20, -8, -4, -1, 3, 9, 15, 24, 0};
    const int queen_pst  [] = {0, 0, 0, 1, 1, 0, 0, 0}; //{0, 0, 0, 0, 0, -5, -10, -20};
    const int king_pst   [] = {-5, -20, -40, -80, -160, -320, -640, -1280};

    const int pst_p      [] = {  0,  0,  0,  0,  0,  0,  0,  0,
                                50, 50, 50, 50, 50, 50, 50, 50,
                                10, 10, 20, 30, 30, 20, 10, 10,
                                 5,  5, 10, 25, 25, 10,  5,  5,
                                 0,  0,  0, 20, 20,  0,  0,  0,
                                 5, -5,-10,  0,  0,-10, -5,  5,
                                 5, 10, 10,-20,-20, 10, 10,  5,
                                 0,  0,  0,  0,  0,  0,  0,  0};

    const int pst_n      [] = {-50,-40,-30,-30,-30,-30,-40,-50,
                               -40,-20,  0,  0,  0,  0,-20,-40,
                               -30,  0, 10, 15, 15, 10,  0,-30,
                               -30,  5, 15, 20, 20, 15,  5,-30,
                               -30,  0, 15, 20, 20, 15,  0,-30,
                               -30,  5, 10, 15, 15, 10,  5,-30,
                               -40,-20,  0,  5,  5,  0,-20,-40,
                               -50,-40,-30,-30,-30,-30,-40,-50};

    const int pst_b      [] = {-20,-10,-10,-10,-10,-10,-10,-20,
                               -10,  0,  0,  0,  0,  0,  0,-10,
                               -10,  0,  5, 10, 10,  5,  0,-10,
                               -10,  5,  5, 10, 10,  5,  5,-10,
                               -10,  0, 10, 10, 10, 10,  0,-10,
                               -10, 10, 10, 10, 10, 10, 10,-10,
                               -10,  5,  0,  0,  0,  0,  5,-10,
                               -20,-10,-10,-10,-10,-10,-10,-20};

    const int pst_r      [] = {  0,  0,  0,  0,  0,  0,  0,  0,
                                 5, 10, 10, 10, 10, 10, 10,  5,
                                -5,  0,  0,  0,  0,  0,  0, -5,
                                -5,  0,  0,  0,  0,  0,  0, -5,
                                -5,  0,  0,  0,  0,  0,  0, -5,
                                -5,  0,  0,  0,  0,  0,  0, -5,
                                -5,  0,  0,  0,  0,  0,  0, -5,
                                 0,  0,  0,  5,  5,  0,  0,  0};

    const int pst_q      [] = {-20,-10,-10, -5, -5,-10,-10,-20,
                               -10,  0,  0,  0,  0,  0,  0,-10,
                               -10,  0,  5,  5,  5,  5,  0,-10,
                                -5,  0,  5,  5,  5,  5,  0, -5,
                                 0,  0,  5,  5,  5,  5,  0, -5,
                               -10,  5,  5,  5,  5,  5,  0,-10,
                               -10,  0,  5,  0,  0,  0,  0,-10,
                               -20,-10,-10, -5, -5,-10,-10,-20};


    for (a = 7; a >= 0; a --) // 
    {
        for (b = 0; b < 8; b ++) // 
        {
            sq = compose (a, b);

            // 
            t = pawn_pst_y [a] + pawn_pst_x [b];

            pst [WP][sq] = t;

            // 
            att = knight_att [sq];
            t = knight_pst [pop_count (att)];

            pst [WN][sq] = t;

            // 
            att = b_att (E, sq);
            t = bishop_pst [pop_count (att)];

            pst [WB][sq] = t;

            // 
            t = 0;
            if (a == 0 || a == 7) t -= 2;
            if (b == 0 || b == 7) t -= 2;

            if (a == 6) t += 5;
            if (sq == 003 || sq == 004) t += 5;

            pst [WR][sq] = t;

            // 
            t = pst [WB][sq] >> 2;
            t += queen_pst [a] * queen_pst [b];

            pst [WQ][sq] = t;

            //  /////////////////////////////////////    + bishop >> 3
            if (a == 0 && (b < 3 || b > 4)) t = 20;
            else
            {
                t = king_pst [a];
            };

            pst [WK][sq] = t;

            //uci_send ("%8i ", pst [WK][sq]);
        };
        //uci_send ("\n");
    };

    //  (  chessprogramming,   )
    for (sq = 0; sq < 64; sq ++)
    {
        pst [WP][sq] = pst_p  [sq^070];
        pst [WN][sq] = pst_n  [sq^070];
        pst [WB][sq] = pst_b  [sq^070];
        pst [WR][sq] = pst_r  [sq^070];
        pst [WQ][sq] = pst_q  [sq^070];
        pst [WK][sq] = pst_km [sq^070];

        //if (!(sq & 07)) uci_send ("\n");

        //uci_send ("%4i ", pst_p [sq]);
    };

    //    
    for (sq = 0; sq < 64; sq ++)
    {
        for (i = 0; i < 12; i += 2)
        {
            pst [i][sq] = pst [i^1][sq^070];
        };
    };
};


void init_attacks ()
{
	int a, b, c, i, dir;
	int st, en;
	U64 bb, att;

	// Knight attacks
	for (a = 0; a < 8; a ++)
	{
		for (b = 0; b < 8; b ++)
		{
			c = compose (a, b);
			bb = I << c;

			att  = (bb & ~(file_a | file_b)) << 6;
			att |= (bb & ~(file_h | file_g)) << 10;
			att |= (bb & ~file_a) << 15;
			att |= (bb & ~file_h) << 17;

			att |= (bb & ~(file_h | file_g)) >> 6;
			att |= (bb & ~(file_a | file_b)) >> 10;
			att |= (bb & ~file_h) >> 15;
			att |= (bb & ~file_a) >> 17;

			knight_att [c] = att;
		};
	};

	// King attacks
	for (a = 0; a < 8; a ++)
	{
		for (b = 0; b < 8; b ++)
		{
			c = compose (a, b);
			bb = I << c;

			att  = (bb & ~file_h) << 1;
			att |= (bb & ~file_a) << 7;
			att |= bb << 8;
			att |= (bb & ~file_h) << 9;

			att |= (bb & ~file_a) >> 1;
			att |= (bb & ~file_h) >> 7;
			att |= bb >> 8;
			att |= (bb & ~file_a) >> 9;

			king_att [c] = att;
		};
	};

	// Pawn attacks
	for (a = 0; a < 8; a ++)
	{
		for (b = 0; b < 8; b ++)
		{
			c = compose (a, b);
			bb = I << c;

			att  = (bb & ~file_a) << 7;
			att |= (bb & ~file_h) << 9;

			wpawn_att [c] = att;

			att  = (bb & ~file_h) >> 7;
			att |= (bb & ~file_a) >> 9;

			bpawn_att [c] = att;
		};
	};

	// Link attacks
	for (a = 0; a < 64; a ++)
	{
		for (b = 0; b < 64; b ++)
		{
			link_att [a][b] = E;

			//uci_send ("%i %i\n", a, b);

			dir = is_on_line (a, b);

			//     
			if (!dir) continue;

			if (a < b)
			{
				st = a;
				en = b;
			}
			else
			{
				st = b;
				en = a;
				dir = -dir;
			};

			for (i = st; i <= en; i += dir)
			{
				link_att [a][b] |= I << i;
			};

			link_att [a][b] &= ~(I << st);
			link_att [a][b] &= ~(I << en);
		};
	};
};

void init_magics ()
{
	int i, j;

	err = 0;

	//   
	if (!GEN_MAGICS && f_magics != NULL)
	{
		for (i = 0; i < 64; i ++)
		{
			for (j = 0; j < (1 << 12); j ++)
			{
				rook_att [i][j] = f_read64 (f_magics);
			};

			rook_m [i].mask  = f_read64 (f_magics);
			rook_m [i].magic = f_read64 (f_magics);
		};

		for (i = 0; i < 64; i ++)
		{
			for (j = 0; j < (1 << 9); j ++)
			{
				bishop_att [i][j] = f_read64 (f_magics);
			};

			bishop_m [i].mask  = f_read64 (f_magics);
			bishop_m [i].magic = f_read64 (f_magics);
		};

		//    
		if (err)
		{
			fclose (f_magics);
			goto exit;
		};

		return;
	};

exit:
	//       
	f_magics = fopen (F_MAGICS, "wb");

	// Rook magics
	uci_send ("const U64 r_magics[64] = {\n");
	for (i = 0; i < 64; i ++) init_rook_magics (i);
	uci_send ("};\n\n");

	// Bishop magics
	uci_send ("const U64 b_magics[64] = {\n");
	for (i = 0; i < 64; i ++) init_bishop_magics (i);
	uci_send ("};\n\n");

	fclose (f_magics);
};

void init_rook_magics (int c)
{
	const int n = 12; //    
	int i, k, x;     // , 
	int fail;       //  
	int bits;      //    
	U64 bb, key, mask, magic;
	U64 x64; //  popcount'

	U64 blk [1 << n]; //  
	U64 atc [1 << n]; //  

	U64 used [1 << n]; //    -

	bb = I << c;
	mask = rmask (c);
	bits = pop_count (mask);

	//uci_send ("%i ", bits);

	//  
	for (i = 0; i < (1 << bits); i ++)
	{
		key = index_to_64 (i, bits, mask);

		blk [i] = key;
		atc [i] = ratt (c, key);

		//print_this_att (key);
		//print_this_att (atc [i]);
	};

	//   

	for (k = 0; k < 100000000; k ++)
	{
		fail = 0;
		magic = rand64_fewbits ();

		//if ( pop_count ( (mask * magic) & (U64)0xFF00000000000000) < 6 ) continue;

		for (i = 0; i < (1 << n); i ++) used[i] = E;

		for (i = 0; !fail && (i < (1 << bits)); i ++)
		{
			bb = blk [i];
			bb *= magic;
			bb >>= (64 - n);

			//   
			if (used [bb] == E) used [bb] = atc [i];
			else if (used [bb] != atc [i]) fail = 1;
		};
		if (!fail) break;
	};

	if (fail) uci_send ("Failed!!!\n");
	else uci_send (" 0x%08X%08X, \n", (int)(magic >> 32), (int)magic);

	//  
	for (i = 0; i < (1 << n); i ++)
	{
		rook_att [c][i] = used [i];
		fprintf (f_magics, "%08X%08X", (int)(used [i] >> 32), (int)used [i]);
	};

	rook_m [c].mask = mask;
	rook_m [c].magic = magic;

	fprintf (f_magics, "%08X%08X", (int)(mask >> 32), (int)mask);
	fprintf (f_magics, "%08X%08X", (int)(magic >> 32), (int)magic);
};



void init_bishop_magics (int c)
{
	const int n = 9;  //    
	int i, k, x;     // , 
	int fail;       //  
	int bits;      //    
	U64 bb, key, mask, magic;
	U64 x64; //  popcount'

	U64 blk [1 << n]; //  
	U64 atc [1 << n]; //  

	U64 used [1 << n]; //    -

	bb = I << c;
	mask = bmask (c);
	bits = pop_count (mask);

	//uci_send ("%i ", bits);

	//  
	for (i = 0; i < (1 << bits); i ++)
	{
		key = index_to_64 (i, bits, mask);

		blk [i] = key;
		atc [i] = batt (c, key);

		//print_this_att (key);
		//print_this_att (atc [i]);
	};

	//   

	for (k = 0; k < 100000000; k ++)
	{
		fail = 0;
		magic = rand64_fewbits ();

		//if ( pop_count ( (mask * magic) & (U64)0xFF00000000000000) < 6) continue;

		for (i = 0; i < (1 << n); i ++) used[i] = E;

		for (i = 0; !fail && (i < (1 << bits)); i ++)
		{
			bb = blk [i];
			bb *= magic;
			bb >>= (64 - n);

			//   
			if (used [bb] == E) used [bb] = atc [i];
			else if (used [bb] != atc [i]) fail = 1;
		};
		if (!fail) break;
	};

	if (fail) uci_send ("Failed!!!\n");
	else uci_send (" 0x%08X%08X, \n", (int)(magic >> 32), (int)magic);

	//  
	for (i = 0; i < (1 << n); i ++)
	{
		bishop_att [c][i] = used [i];
		fprintf (f_magics, "%08X%08X", (int)(used [i] >> 32), (int)used [i]);
	};

	bishop_m [c].mask = mask;
	bishop_m [c].magic = magic;

	fprintf (f_magics, "%08X%08X", (int)(mask >> 32), (int)mask);
	fprintf (f_magics, "%08X%08X", (int)(magic >> 32), (int)magic);
};

















U64 rand64 ()
{
    U64 u1, u2, u3, u4;

	u1 = (U64)(rand()) & 0xFFFF;
	u2 = (U64)(rand()) & 0xFFFF;
	u3 = (U64)(rand()) & 0xFFFF;
	u4 = (U64)(rand()) & 0xFFFF;

	return u1 | (u2 << 16) | (u3 << 32) | (u4 << 48);
}

U64 rand64_fewbits()
{
	return rand64() & rand64() & rand64();
};

U64 rmask (int sq)
{
	U64 result = E;
	int rk, fl, r, f;

	rk = hori (sq);
	fl = vert (sq);

	for (r = rk + 1; r <= 6; r ++) result |= I << compose(r, fl);
	for (r = rk - 1; r >= 1; r --) result |= I << compose(r, fl);
	for (f = fl + 1; f <= 6; f ++) result |= I << compose(rk, f);
	for (f = fl - 1; f >= 1; f --) result |= I << compose(rk, f);

	return result;
};

U64 bmask (int sq)
{
	U64 result = E;
	int rk, fl, r, f;

	rk = hori (sq);
	fl = vert (sq);

	for (r = rk + 1, f = fl + 1; r <= 6 && f <= 6; r ++, f ++) result |= I << compose(r, f);
	for (r = rk + 1, f = fl - 1; r <= 6 && f >= 1; r ++, f --) result |= I << compose(r, f);
	for (r = rk - 1, f = fl + 1; r >= 1 && f <= 6; r --, f ++) result |= I << compose(r, f);
	for (r = rk - 1, f = fl - 1; r >= 1 && f >= 1; r --, f --) result |= I << compose(r, f);

	return result;
};

U64 qmask (int sq)
{
	U64 result = E;
	int rk, fl, r, f;

	rk = hori (sq);
	fl = vert (sq);

	for (r = rk + 1; r <= 7; r ++) result |= I << compose(r, fl);
	for (r = rk - 1; r >= 0; r --) result |= I << compose(r, fl);
	for (f = fl + 1; f <= 7; f ++) result |= I << compose(rk, f);
	for (f = fl - 1; f >= 0; f --) result |= I << compose(rk, f);

	for (r = rk + 1, f = fl + 1; r <= 7 && f <= 7; r ++, f ++) result |= I << compose(r, f);
	for (r = rk + 1, f = fl - 1; r <= 7 && f >= 0; r ++, f --) result |= I << compose(r, f);
	for (r = rk - 1, f = fl + 1; r >= 0 && f <= 7; r --, f ++) result |= I << compose(r, f);
	for (r = rk - 1, f = fl - 1; r >= 0 && f >= 0; r --, f --) result |= I << compose(r, f);

	return result;
};

U64 ratt (int sq, U64 block)
{
	U64 bb, result = E;
	int rk, fl, r, f;

	rk = hori (sq);
	fl = vert (sq);

	for (r = rk + 1; r <= 7; r ++)
	{
		bb = I << compose(r, fl);
		result |= bb;
		if (block & bb) break;
	};
	for (r = rk - 1; r >= 0; r --)
	{
		bb = I << compose(r, fl);
		result |= bb;
		if (block & bb) break;
	};
	for (f = fl + 1; f <= 7; f ++)
	{
		bb = I << compose(rk, f);
		result |= bb;
		if (block & bb) break;
	};
	for (f = fl - 1; f >= 0; f --)
	{
		bb = I << compose(rk, f);
		result |= bb;
		if (block & bb) break;
	};

	return result;
};

U64 batt (int sq, U64 block)
{
	U64 bb, result = E;
	int rk, fl, r, f;

	rk = hori (sq);
	fl = vert (sq);

	for (r = rk + 1, f = fl + 1; r <= 7 && f <= 7; r ++, f ++)
	{
		bb = I << compose(r, f);
		result |= bb;
		if (block & bb) break;
	};
	for (r = rk + 1, f = fl - 1; r <= 7 && f >= 0; r ++, f --)
	{
		bb = I << compose(r, f);
		result |= bb;
		if (block & bb) break;
	};
	for (r = rk - 1, f = fl + 1; r >= 0 && f <= 7; r --, f ++)
	{
		bb = I << compose(r, f);
		result |= bb;
		if (block & bb) break;
	};
	for (r = rk - 1, f = fl - 1; r >= 0 && f >= 0; r --, f --)
	{
		bb = I << compose(r, f);
		result |= bb;
		if (block & bb) break;
	};

	return result;
};

U64 index_to_64 (int index, int bits, U64 m)
{
	int i;
	U64 t, result = E;

	for (i = 0; i < bits; i++)
	{
		t = LSB (m);
		if (index & (1 << i)) result |= t;
		m ^= t;
	};

	return result;
};

void print_pv (int len)
{
	int i;
	U64 key;
	int mv;

	for (key = S->pv_start, i = key & (PV_SIZE - 1); key != NULL;)
	{
	    i = key & (PV_SIZE - 1);
	    while (pv[i].mv && key != pv[i].key) i = (i < PV_SIZE - 1) ? (i + 1) : 0;

	    mv = pv [i].mv;
	    if (!mv) break;

		uci_send ("%s ", move_to_str (mv));

		key = pv[i].key_new;
		if (key == S->pv_start) break;
	};

	uci_send ("\n");
};

void sprint_pv (int len, char* pv_line)
{
	int   i, j;
	U64   key;
	char* move;
	char  ch;
	int   mv;
	U64   keys [128];
	int   curr_key = 0;

    for (key = S->pv_start, i = key & (PV_SIZE - 1); key != NULL;)
	{
	    i = key & (PV_SIZE - 1);
	    while (pv[i].mv && key != pv[i].key) i = (i < PV_SIZE - 1) ? (i + 1) : 0;

        mv = pv [i].mv;

        if (curr_key, 0)
        {
            for (j = 0; j < curr_key - 1; j ++)
            {
                if (key == keys[j])
                {
                    mv = 0;
                    break;
                };
            };
        };

	    if (!mv) break;

		move = move_to_str (mv);

		while (ch = *move++) *pv_line++ = ch;

		key = pv[i].key_new;
		if (key == S->pv_start) break;

		*pv_line++ = ' ';

		keys [curr_key ++] = key;
		if (curr_key > 40) break;
	};

	*pv_line = 0;
};

U64 f_read64 (FILE* file)
{
	int t;
	U64 bb, b;
	U32 p;

	t = fscanf (file, "%8x", &p);
	b = p;
	bb = b << 32;

	if (t == -1)
	{
		err = 1;
		return U;
	};

	t = fscanf (file, "%8x", &p);
	b = p;
	bb += b & 0xFFFFFFFF;

	if (t == -1)
	{
		err = 1;
		return U;
	};

	return bb;
};

void print_this_att (U64 at)
{
	int a, b, c;
	U64 p;
	U64 t = 0;

	for (a = 0; a <= 7; a ++)
	{
		uci_send ("\n");
		for (b = 0; b <= 7; b ++)
		{
			c = compose ((7 - a), b);
			t = I << c;

			p = at & t;

			if (p) uci_send ("#");
			else   uci_send (".");
		};
	};
	uci_send ("\n");
};

int print_move (int mv)
{
	int fl, ep;
	int _from, _to, _flags, _piece;

	_flags = (mv & 0x000000FF);
    _to    = (mv & 0x0000FF00) >> 8;
    _from  = (mv & 0x00FF0000) >> 16;
    _piece = (mv & 0xFF000000) >> 24;

	//if (_from = -1) return 0;

	uci_send (" %c", piece_array [_piece]);
	uci_send ("%c%c", 'a' + (vert (_from)), '1' + (hori (_from)));

	ep = is_ep(_flags);
	fl = is_capture(_flags) || ep;
	if (fl) uci_send ("x");

	uci_send ("%c%c", 'a' + (vert (_to)), '1' + (hori (_to)));

	if (!fl) uci_send (" ");
	if (ep) uci_send (" ep");

	//uci_send ("\n");

	return 0;
};

void shell_sort (int * a, int * b, int num)
{
    register int i, j, k, inc;
    int x, y;

    int arr[] = {9, 5, 3, 2, 1};

    for (k = 0; k < 5; k ++)
    {
        inc = arr[k];
        for (i = inc; i < num; ++ i)
        {
            x = a[i];
            y = b[i];

            for (j = i - inc; (x > a[j]) && (j >= 0); j -= inc)
            {
                a[j + inc] = a[j];
                b[j + inc] = b[j];
            };

            a[j + inc] = x;
            b[j + inc] = y;
        };
    };
};


inline U64 _r_att (U64 a, int sq)
{
    a &= rook_m[sq].mask;
    a *= rook_m[sq].magic;
    a >>= (64 - 12);

    return rook_att [sq][a];
};

inline U64 _b_att (U64 a, int sq)
{
    a &= bishop_m[sq].mask;
    a *= bishop_m[sq].magic;
    a >>= (64 -  9);

    return bishop_att [sq][a];
};

inline int is_on_line (int a, int b)
{
    int h, v;

    h = hori (b) - hori (a);
    v = vert (b) - vert (a);

    if (h > 0)
    {
        if (v > 0) return (abs (h) == abs (v)) ? 9 : 0;
        if (v < 0) return (abs (h) == abs (v)) ? 7 : 0;
        return 8;
    };

    if (h < 0)
    {
        if (v > 0) return (abs (h) == abs (v)) ? -7 : 0;
        if (v < 0) return (abs (h) == abs (v)) ? -9 : 0;
        return -8;
    };

    if (v > 0) return 1;
    if (v < 0) return -1;

    return 0;
};

int _pop_count    (U64 x)
{
    //return pop_count_few (x);
    static const U64 k1 = C64(0x5555555555555555); /*  -1/3   */
    static const U64 k2 = C64(0x3333333333333333); /*  -1/5   */
    static const U64 k4 = C64(0x0f0f0f0f0f0f0f0f); /*  -1/17  */
    static const U64 kf = C64(0x0101010101010101); /*  -1/255 */

    x =  x       - ((x >> 1)  & k1); /* put count of each 2 bits into those 2 bits */
    x = (x & k2) + ((x >> 2)  & k2); /* put count of each 4 bits into those 4 bits */
    x = (x       +  (x >> 4)) & k4 ; /* put count of each 8 bits into those 8 bits */
    x = (x * kf) >> 56;              /* returns 8 most significant bits of x + (x<<8) + (x<<16) + (x<<24) + ...  */
    return (int) x;
};

int _pop_count_few (U64 x)
{
   int count = 0;

   while (x)
   {
       count ++;
       x &= x - 1; // reset LS1B
   }
   return count;
};


/*
// CountBits()

inline uint CountBits(BitBoard bb) {
  unsigned w = unsigned(bb >> 32), v = unsigned(bb);
  v -= (v >> 1) & 0x55555555;
  w -= (w >> 1) & 0x55555555;
  v = ((v >> 2) & 0x33333333) + (v & 0x33333333);
  w = ((w >> 2) & 0x33333333) + (w & 0x33333333);
  v = ((v >> 4) + v) & 0x0F0F0F0F;
  v += (((w >> 4) + w) & 0x0F0F0F0F);
  v *= 0x01010101;
  return int(v >> 24);
}
*/// From Bison
